#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <fnmatch.h>

#include "lbx.h"

static const char *progname;
#define errmsg(fmt, ...) (\
	fprintf(stderr, "%s: " fmt, progname, __VA_ARGS__)\
)

enum {
	MODE_NONE,
	MODE_LIST,
	MODE_EXTRACT,
};

int filematch(char **argv, const char *name)
{
	int rc, i;

	for (i = 0; argv[i]; i++) {
		switch(fnmatch(argv[i], name, 0)) {
		case 0:
			return 0;
		case FNM_NOMATCH:
			break;
		default:
			errmsg("error matching glob: %s.\n", argv[i]);
			return 1;
		}
	}

	return i ? -1: 0;
}

int list(FILE *f, const char *name, int verbose, char **argv) {
	LBX *lbx;
	size_t nfiles;
	unsigned int i;

	lbx = lbx_fopen(f, name);
	if (!lbx) {
		errmsg("failed to open archive: %s.\n", lbx_strerror());
		return EXIT_FAILURE;
	}

	nfiles = lbx_numfiles(lbx);
	if (verbose) {
		printf("Files in archive: %zu\n", nfiles);
	}

	for (i = 0; i < nfiles; i++) {
		struct lbx_statbuf stat;

		lbx_stat(lbx, i, &stat);

		switch (filematch(argv, stat.name)) {
		case -1: continue;
		case  0: break;
		default: goto err;
		}

		printf("%s", stat.name);
		if (verbose) {
			printf(" size=%zu bytes", stat.size);
		}

		putchar('\n');
	}

	lbx_close(lbx);
	return EXIT_SUCCESS;
err:
	lbx_close(lbx);
	return EXIT_FAILURE;
}

int extract(FILE *f, const char *name, int verbose, char **argv) {
	LBX *lbx;
	size_t nfiles;
	unsigned int i;

	lbx = lbx_fopen(f, name);
	if (!lbx) {
		errmsg("failed to open archive: %s.\n", lbx_strerror());
		return EXIT_FAILURE;
	}

	nfiles = lbx_numfiles(lbx);
	if (verbose) {
		printf("Files in archive: %zu\n", nfiles);
	}

	for (i = 0; i < nfiles; i++) {
		struct lbx_statbuf stat;
		size_t rc;
		FILE *of;
		int j;

		lbx_stat(lbx, i, &stat);

		switch (filematch(argv, stat.name)) {
		case -1: continue;
		case  0: break;
		default: goto err;
		}

		of = fopen(stat.name, "wbx");
		if (!of) {
			errmsg("failed to create output file %s: %m.\n",
			                                     stat.name);
			goto err;
		}

		if (verbose) printf("extracting %s...\n", stat.name);
		rc = lbx_extract(lbx, i, of);
		if (rc < stat.size) {
			if (rc == 0) remove(stat.name);
			errmsg("error extracting %s: %s.\n",
			       stat.name, lbx_strerror());
			goto err;
		}
		if (verbose) printf("wrote %zu bytes.\n", rc);
	}

	lbx_close(lbx);
	return EXIT_SUCCESS;
err:
	lbx_close(lbx);
	return EXIT_FAILURE;
}

int main(int argc, char **argv)
{
	int mode = MODE_NONE, verbose = 0;
	const char *name = "stdin";
	FILE *f  = stdin;
	int opt;

	static const char         *sopts   = "lxf:i:v";
	static const struct option lopts[] = {
		{ "list",    0, NULL, 'l' },
		{ "extract", 0, NULL, 'x' },

		{ "file",    1, NULL, 'f' },
		{ "index",   1, NULL, 'i' },

		{ "verbose", 0, NULL, 'v' },

		{ 0 }
	};

	progname = "lbxtool"; /* argv[0]; */
	while ((opt = getopt_long(argc, argv, sopts, lopts, NULL)) != -1) {
		switch(opt) {
		case 'l':
			mode = MODE_LIST;
			break;
		case 'x':
			mode = MODE_EXTRACT;
			break;
		case 'f':
			if (strcmp(optarg, "-") == 0)
				break;

			name = strrchr(optarg, '/');
			name = name ? name+1 : optarg;

			f = fopen(optarg, "rb");
			if (!f) {
				errmsg("failed to open file %s: %m\n", optarg);
				return EXIT_FAILURE;
			}
			break;
		case 'i':
			/* FIXME: Add index file support. */
			break;
		case 'v':
			verbose = 1;
			break;
		default:
			return EXIT_FAILURE;
		}
	}

	switch (mode) {
	case MODE_LIST:
		return list(f, name, verbose, &argv[optind]);
	case MODE_EXTRACT:
		return extract(f, name, verbose, &argv[optind]);
	}

	fprintf(stderr, "%s: you must specify a mode.\n", progname);
	return EXIT_FAILURE;
}
